#include <stdlib.h>

#include <pqueue.h>

/*
Ken er implementeret som en enkelt hktet cirkulr liste af
'entry_t' strukturer. Listen er sorteret i henhold til prioritet.

Hndtaget 'pq' som brugeren angiver peger (pq->e) p den entry med laveste
prioritet. Den cirkulre udformining gr at 'pq->e->next' vil vre
den entry med hjst prioritet.

                 pq->e
                   |
 8 --> 5 --> 4 --> 2 --
 ^____________________|

*/


struct entry {
    struct entry *next;
    void *data;
    unsigned int pri;
};

/*
Indsttelse sker ved at sge igennem listen og finde den rigtige plads.
Kompleksitet O(n).
*/
int pqueue_insert(pqueue_t *pq, unsigned int pri, void *data) {
    entry_t *e;

    e = malloc(sizeof(entry_t));
    if (e == NULL) {
        return -1;
    }

    e->data = data;
    e->pri = pri;

    if (pq->e == NULL) {
        pq->e = e;
        e->next = e;
    } else if (pri <= pq->e->pri) {
        e->next = pq->e->next;
        pq->e->next = e;
        pq->e = e;
    } else {
        entry_t *i = pq->e;

        while (pri <= i->next->pri) {
            i = i->next;
        }
        e->next = i->next;
        i->next = e;
    }
    return 0;
}

/*
Udtagelse sker ved at fjene den entry med hjst prioritet og stte den med
lavest til at pege p den anden hjest.
*/
void *pqueue_remove(pqueue_t *pq) {
    void *d;
    entry_t *e;

    if (pq->e == NULL) {
        return NULL;
    }

    e = pq->e->next;
    d = e->data;

    pq->e->next = e->next;

    if (e == pq->e) {
        pq->e = NULL;
    }
    free(e);
    return d;
}
