#include "wqueue.h"
#include <stdio.h>
#include <stdlib.h>

/*
Indstter et stykke arbejde i form af en funktion 'func' og det
data 'data' den skal kaldes med. Prioriteten 'pri' angiver
arbejdets prioritet. 'wq' peger p en instans af arbejdsken.

Lykkedes indsttelsen skal funktionen returnerer '0'. Ved fejl '-1'.
*/
int wqueue_insert(wqueue_t *wq, unsigned int pri, work_f func, void *data)
{
  
  if(data == NULL || pri <= 0) /* vi tillader ikke at indstte tomme elementer, eller elementer med pri < 1 */
    exit(-1); 

  if(wq->data == NULL) /* det er en tom liste og data bliver det frste element */
    {  
      wq->data = data;
      wq->pri = pri;
      wq->func = func;
      wq->prev = wq; /* pga remove skal wq pege p sig selv hvis det er det eneste element i ken */ 
    }

  else /* ken er ikke tom, og vi skal indstte data det rigtige sted...  */
    {
      int temp_pri;
      temp_pri = wq->pri;
      wqueue_t *temp_q;
      temp_q = wq;

      while (temp_pri < pri && temp_q->next != NULL)
	{
	  temp_q = temp_q->next;
	  temp_pri = temp_q->pri;
	}

      /* vi opretter en ny plads i ken og stter data=data og pri=pri */
      wqueue_t *new = malloc(sizeof(wqueue_t));
      if (new == NULL)
	{exit(-1);}
      else
	{
	  new->data = data;
	  new->pri = pri;
	  new->func = func;
	}

      /* hvis data skal indsttes sidst i ken (data har hjest prioritet) */
      if(temp_q->next == NULL)
	{
	  temp_q->next = new;
	  new->next = NULL;
	  new->prev = temp_q;
	  /*
	    da elementet med hjeste pri altid vil vre det sidste
	    element i ken, stter vi frste elements prev til at
	    pege p det sidste element, s vi ikke skal kigge hele
	    ken igennem nr vi skal fjerne et element fra ken
	  */
	  wq->prev = new;
	}
      

      /* Hvis data skal sttes ind imellem andre elementer i ken */

      else
	{
	  new->prev = temp_q->prev;	
	  new->next = temp_q;
	  if(temp_q == head) /* hvis det er et nyt frste element (nyt head -> lavest pri) */
	    {
	      temp_q->prev = new;
	      head = new;
	    }
	  else
	    {
	      (temp_q->prev)->next = new; 
	      temp_q->prev = new;
	    } 
	}

    }
  return 0;
 
}


/*
Udfrer et stykke arbejde fra ken.

Hvis ken er tom returneres '0', ellers '1'.

*/
int wqueue_run(wqueue_t *wq)
{
  if(wq->data == NULL)
    {
      return 0;
    }

  else
    { 
      
      if(wq->prev == wq) /* dette er det eneste element i ken */
	{
	  /* udtrk funktion og data pegere */

	  void *temp_data;
	  work_f temp_work_f;
	  temp_data =  wq->data;
	  temp_work_f = wq->func;

	  /* Da ken nu er tom, "nulstiller" vi hovedet af ken */
	  wq->pri = 0;
	  wq->data = NULL;
	  wq->func = NULL;
	  wq->next = NULL;
	  wq->prev = NULL;

	  /* udfr arbejdet */
	  (*temp_work_f)(temp_data);
	}

      else /* "normale" tilflde */
	{
	  wqueue_t *temp_q;	  
	  temp_q = wq->prev;
	  /* frste elements pointer til sidste element flyttes til nstsidste element */
	  wq->prev = temp_q->prev;
	  /* det nstsidste element er nu det sidste element*/
	  (temp_q->prev)->next = NULL;
	  
	  /* udfr arbejdet */
	  (*(temp_q->func))(temp_q->data);

	}

      return 1;
    }
}

wqueue_t *init_queue()
{

  head = malloc(sizeof(wqueue_t));
  if (head == NULL)
    exit(-1);
  head->pri  = 0;
  head->func = NULL;
  head->data = NULL;
  head->next = NULL;
  head->prev = NULL;
  return head;
  
}

